/* cfgfile.c
 * Part of ziproxy package.
 * 
 * Copyright (c)2003-2004 Juraj Variny<variny@naex.sk>
 * Copyright (c)2005-2006 Daniel Mealha Cabrita
 *
 * Released subject to GNU General Public License.
 * 
 * Configuration file parsing functions.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <confuse.h>


#ifdef CONFUSEv2
cfg_bool_t MSIETest, DoGzip, UseContentLength, ModifySuffixes, AllowLookCh, ProcessJPG, ProcessPNG, ProcessGIF, PreemptNameRes, PreemptNameResBC, TransparentProxy, ConventionalProxy, ProcessHTML, ProcessCSS, ProcessJS, ProcessHTML_CSS, ProcessHTML_JS, ProcessHTML_tags, ProcessHTML_text, ProcessHTML_PRE, ProcessHTML_NoComments;
#else
cfgbool_t MSIETest, DoGzip, UseContentLength, ModifySuffixes, AllowLookCh, ProcessJPG, ProcessPNG, ProcessGIF, PreemptNameRes, PreemptNameResBC, TransparentProxy, ConventionalProxy, ProcessHTML, ProcessCSS, ProcessJS, ProcessHTML_CSS, ProcessHTML_JS, ProcessHTML_tags, ProcessHTML_text, ProcessHTML_PRE, ProcessHTML_NoComments;
#endif

int Port, NextPort, NetdTimeout, ZTimeout, MaxSize, PreemptNameResMax, MinTextStream;;
int ImageQ[4];
float JP2Rate[4];
char *WhereZiproxy, *ServHost, *OnlyFrom, *LogFileName, *NextProxy, *ServUrl;
char **LogPipe, **Compressible;
char *CustomError400, *CustomError404, *CustomError408, *CustomError500, *CustomError503;
char *Address;
char *AccessLogFileName;
char *PasswdFile;
char **PasswdEntries;
unsigned long PasswdNumEntries;

#ifndef DefaultCfgLocation
char DefaultCfgLocation[] = "/etc/ziproxy.conf";
#endif

#ifdef CONFUSEv2
static cfg_opt_t opts[] = {
#else
static cfgopt_t opts[] = {
#endif
	CFG_SIMPLE_STR("WhereZiproxy", &WhereZiproxy),
	CFG_SIMPLE_STR("ViaServer", &ServHost),
	CFG_SIMPLE_STR("ServerUrl", &ServUrl),
        CFG_SIMPLE_INT("Port", &Port),
	CFG_SIMPLE_STR("OnlyFrom", &OnlyFrom),
	CFG_SIMPLE_STR("LogFile", &LogFileName),
	CFG_STR("LogPipe", 0 , CFGF_LIST),	
	CFG_SIMPLE_STR("NextProxy", &NextProxy),
        CFG_SIMPLE_INT("NextPort", &NextPort),
	CFG_SIMPLE_INT("NetdTimeout", &NetdTimeout),
	CFG_SIMPLE_INT("ZiproxyTimeout", &ZTimeout),
	CFG_SIMPLE_BOOL("MSIETest", &MSIETest),
	CFG_BOOL("Gzip", cfg_true, 0),
	CFG_STR("Compressible", 0 , CFGF_LIST),	
	CFG_SIMPLE_INT("MaxSize", &MaxSize),
	CFG_SIMPLE_BOOL("UseContentLength", &UseContentLength),
	CFG_INT("ImageQuality", 0, CFGF_LIST),
	CFG_FLOAT("JP2Rate", 0, CFGF_LIST),
	CFG_BOOL("AllowLookChange", cfg_false, 0),
	CFG_SIMPLE_BOOL("ProcessJPG", &ProcessJPG),
	CFG_SIMPLE_BOOL("ProcessPNG", &ProcessPNG),
	CFG_SIMPLE_BOOL("ProcessGIF", &ProcessGIF),
	CFG_SIMPLE_BOOL("PreemptNameRes", &PreemptNameRes),
	CFG_SIMPLE_INT("PreemptNameResMax", &PreemptNameResMax),
	CFG_SIMPLE_BOOL("PreemptNameResBC", &PreemptNameResBC),
	CFG_SIMPLE_BOOL("TransparentProxy", &TransparentProxy),
	CFG_SIMPLE_BOOL("ConventionalProxy", &ConventionalProxy),
	CFG_SIMPLE_STR("CustomError400", &CustomError400),
	CFG_SIMPLE_STR("CustomError404", &CustomError404),
	CFG_SIMPLE_STR("CustomError408", &CustomError408),
	CFG_SIMPLE_STR("CustomError500", &CustomError500),
	CFG_SIMPLE_STR("CustomError503", &CustomError503),
	CFG_SIMPLE_STR("Address", &Address),
	CFG_SIMPLE_STR("AccessLogFileName", &AccessLogFileName),
	CFG_SIMPLE_STR("PasswdFile", &PasswdFile),
	CFG_SIMPLE_INT("MinTextStream", &MinTextStream),
	CFG_SIMPLE_BOOL("ProcessHTML", &ProcessHTML),
	CFG_SIMPLE_BOOL("ProcessCSS", &ProcessCSS),
	CFG_SIMPLE_BOOL("ProcessJS", &ProcessJS),
	CFG_SIMPLE_BOOL("ProcessHTML_CSS", &ProcessHTML_CSS),
	CFG_SIMPLE_BOOL("ProcessHTML_JS", &ProcessHTML_JS),
	CFG_SIMPLE_BOOL("ProcessHTML_tags", &ProcessHTML_tags),
	CFG_SIMPLE_BOOL("ProcessHTML_text", &ProcessHTML_text),
	CFG_SIMPLE_BOOL("ProcessHTML_PRE", &ProcessHTML_PRE),
	CFG_SIMPLE_BOOL("ProcessHTML_NoComments", &ProcessHTML_NoComments),
#ifdef JP2K
	CFG_BOOL("ModifySuffixes", cfg_false, 0),
#else
	CFG_BOOL("ModifySuffixes", cfg_false, 0),
#endif	
	CFG_STR("Nameservers", 0 , CFGF_LIST),
	CFG_END()
	};

int ReadCfgFile(char * cfg_file)
{
	cfg_t *cfg;
	int status, i, n;
	
	Port = NextPort = 8080;
	NetdTimeout = 240;
	ZTimeout = 90;
	MaxSize = 0;
	PreemptNameResMax = 50;
	MinTextStream = 20000;
	
	JP2Rate[0] = 0.04;
	JP2Rate[1] = 0.04;
	JP2Rate[2] = 0.04;
	JP2Rate[3] = 0.04;

	ImageQ[0] = 17;
	ImageQ[1] = 20;
	ImageQ[2] = 23;
	ImageQ[3] = 25;
	
#ifdef JP2K
	ModifySuffixes = cfg_true;
#else
	ModifySuffixes = cfg_false;
#endif	
	MSIETest = PreemptNameRes = cfg_false;
	DoGzip = UseContentLength = ProcessJPG = ProcessPNG = ProcessGIF = cfg_true;
	ProcessHTML = ProcessCSS = ProcessJS = cfg_false;
	ProcessHTML_CSS = ProcessHTML_JS = ProcessHTML_tags = ProcessHTML_text = ProcessHTML_PRE = ProcessHTML_NoComments = cfg_true;
	AllowLookCh = PreemptNameResBC = TransparentProxy = cfg_false;
	ConventionalProxy = cfg_true;
	WhereZiproxy = ServHost = ServUrl = OnlyFrom = LogFileName = NextProxy = NULL;
	LogPipe = Compressible = NULL;
	CustomError400 = CustomError404 = CustomError408 = CustomError500 = CustomError503 = NULL;
	Address = NULL;
	AccessLogFileName = NULL;
	PasswdFile = NULL;

	cfg = cfg_init(opts, CFGF_NOCASE);
	status = cfg_parse(cfg, cfg_file);
	if(!status){
		n = cfg_size(cfg, "LogPipe");
		if(n){
			LogPipe = malloc((n + 1)*sizeof(char*));
			for(i = 0; i < n; i++) 
				LogPipe[i] = strdup(cfg_getnstr(cfg, "LogPipe", i));
    			
			LogPipe[i] = NULL;
		}
		
		n = cfg_size(cfg, "Compressible");
		if(n){
			Compressible = malloc((n + 1)*sizeof(char*));
			for(i = 0; i < n; i++) 
				Compressible[i] = strdup(cfg_getnstr(cfg, "Compressible", i));
    			
			Compressible[i] = NULL;
		}
		
		n = cfg_size(cfg, "ImageQuality");
		if(n == 4)
			for(i = 0; i < n; i++)
			{
				long q = cfg_getnint(cfg, "ImageQuality", i);

				if((q >= -100) && (q <= 100))
					ImageQ[i] = q;
				else
					fprintf(stderr, "Configuration error: \
							Image quality %ld not in <-100, 100> range.\n\
							Continuing with default value.\n", q);
			}
		else if (n != 0)
			fprintf(stderr,"Configuration error: \
					ImageQuality option must have exactly 4 entries.\n\
					Continuing with default values.\n");
		
		n = cfg_size(cfg, "Nameservers");
		if (n) {
			struct sockaddr_in nsaddr;

			if (n > MAXNS)
				fprintf(stderr, "Configuration error: "
						"You can not specify more than %d Nameservers. "
						"Using nameserver 1-%d.\n", MAXNS, MAXNS);

			nsaddr.sin_family = AF_INET;
			nsaddr.sin_port = htons(53);

			res_init();
			_res.nscount = 0;

			for(i = 0; i < n && i < MAXNS; i++) {
				inet_aton(cfg_getnstr(cfg, "Nameservers", i), &nsaddr.sin_addr);
				_res.nsaddr_list[_res.nscount++] = nsaddr;
			}
		}

		AllowLookCh=cfg_getbool(cfg,"AllowLookChange");
		
		n = cfg_size(cfg, "JP2Rate");
			if(n == 4)
				for(i = 0; i < n; i++)
				{
					float q = cfg_getnfloat(cfg, "JP2Rate", i);

					if((q >= 0.0) && (q <= 1.0))
						JP2Rate[i] = q;
					else
						fprintf(stderr, "Configuration error: \
								JP2 compression rate %f not in <0.0, 1.0> range.\n\
								Continuing with default value.\n", q);
				}
			else if (n != 0)
				fprintf(stderr,"Configuration error: \
						JP2Rate option must have exactly 4 entries.\n\
						Continuing with default values.\n");
		DoGzip=cfg_getbool(cfg,"Gzip");
		ModifySuffixes=cfg_getbool(cfg,"ModifySuffixes");
		if ( !ReadPasswdFile(PasswdFile) )
		    PasswdFile=NULL;
		    
	}
	cfg_free(cfg);
	return status;
}
/*
 * Fills in the values of ImageQuality and JP2Rate according to image dimensions. 
 * Returns index used to get the values ranging from 0 (smallest images)
 * to 3 (largest images).
 */
int getRateQuality(int width, int height, int *quality, float *rate)
{
	int try;
	long i = width * height;
	
	if (i < 5000)
		try = 0;
	else if ((i < 50000)|| (width < 150) || (height < 150))
		try = 1;
	else if (i < 250000)
		try = 2;
	else try = 3;

	*rate = JP2Rate[try];
	*quality = ImageQ[try];
	return try;
}

int ReadPasswdFile(char * pswd_file) {
	FILE *txtfile;
	unsigned long txtfilesize;
	PasswdNumEntries=0;
	char *line;

	line = malloc(200);
	if (pswd_file != NULL){
		if (txtfile = fopen (pswd_file, "r")){
			fseek (txtfile, 0, SEEK_END);
			txtfilesize = ftell (txtfile);
			fseek (txtfile, 0, SEEK_SET);
			PasswdEntries = malloc (txtfilesize + 1);			
			    while ( fgets(line,128,txtfile) ) {
				line[strlen(line)-1]=0;
				if (strlen(line)>3 && strchr(line,':')) {
				    PasswdEntries[PasswdNumEntries]=strdup(line);
				    PasswdNumEntries++;
				}
			    }
			    fclose (txtfile);
		}
	}
	free(line);
	return(PasswdNumEntries);
}
