/* http.c
 * Part of ziproxy package.
 * 
 * Copyright (c)2003-2004 Juraj Variny<variny@naex.sk>
 * Copyright (c)2005-2007 Daniel Mealha Cabrita
 *
 * Released subject to GNU General Public License.
 * 
 * HTTP communication functions.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define SRC_HTTP_C

#include "http.h"
#include "image.h"
#include "cfgfile.h"
#include "htmlmodify.h"
#include "htmlopt.h"
#include <string.h>
#include <stdlib.h>

int is_sending_data = 0;

static char line[MAX_LINELEN];

//Local forwards.

static int check_trim( char* line ); 
int forward_content(http_headers *hdr, FILE *from, FILE *to);
static void read_content(http_headers *hdr, FILE *sockrfp, char ** inbuf, int *inlen);
static void clean_hdr(char* ln);

static const char cd64[]="|$$$}rstuvwxyz{$$$$$$$>?@ABCDEFGHIJKLMNOPQRSTUVW$$$$$$XYZ[\\]^_`abcdefghijklmnopq";
char *base64_decode(char *instr);

// close( sockfd );
void proxy_http (http_headers *client_hdr, FILE* sockrfp, FILE* sockwfp, const char *client_addr, struct timeval *giventime)
{
	http_headers *serv_hdr;
	long i;
	char *tempp;
	char *inbuf, *outbuf;
	int inlen, outlen, status = 0;
	int original_size;	
	
	is_sending_data = 0;
	
#ifdef JP2K
	if(ModifySuffixes){
		client_hdr->proto[7] = '1';
		add_header(client_hdr, "Keep-Alive: 300"); //ZTimeout instead?
		add_header(client_hdr, "Connection: Keep-Alive");
		if(!find_header("Host", client_hdr)){
			snprintf(line, sizeof(line), "Host: %s", client_hdr->host);
			add_header(client_hdr, line);
		}
	}else
#endif
	add_header(client_hdr, "Connection: close");

	// Send request
	logputs("Headers sent to server:");
	
	send_headers_to(sockwfp, client_hdr);

	// If there's content, forward that too
	if(client_hdr->content_length > 0)
		forward_content(client_hdr, stdin, sockwfp);

	logdifftime("Connecting, forwarding headers");
	
	// Get the response
	logputs("In Headers:");
	serv_hdr = get_response_headers(sockrfp);
	
	decide_what_to_do(client_hdr, serv_hdr);

	//log so far
	logprintf("Image = %d, Modify = %d, Chunked = %d\n",
		serv_hdr->type, (serv_hdr->flags & DO_MODIFY) != 0, 
		(serv_hdr->where_chunked > 0));

	if(DoGzip) logprintf("WillGZip = %d, Compress = %d\n",
		(client_hdr->flags & H_WILLGZIP) != 0, 
		(serv_hdr->flags & DO_COMPRESS) != 0);

	//if no data requested only forward header and exit
	if (strcasecmp(client_hdr->method, "HEAD") == 0) {
       		logputs("Forwarding header only.");
		add_header(serv_hdr, "Connection: close");
		add_header(serv_hdr, "Proxy-Connection: close");

		send_headers_to(stdout, serv_hdr);
		fflush(stdout);

		access_log (client_hdr, 0, 0, client_addr, giventime);
		return;
	}

	// not workable, stream data and exit	
	if ((MaxSize && (serv_hdr->content_length > MaxSize)) ||
			!(serv_hdr->flags & DO_COMPRESS ) && 
			!(serv_hdr->flags & DO_MODIFY) &&
			!serv_hdr->type){
		is_sending_data = 1;
		if (serv_hdr->flags & H_SIMPLE_RESPONSE){
			logputs("Sending HTTP/1.0 Simple Response.");
		  	fputs(serv_hdr->hdr[0], stdout); // first few bytes of a simple response
		}else {
			logputs("Forwarding header and streaming data.");
			add_header(serv_hdr, "Connection: close");
			add_header(serv_hdr, "Proxy-Connection: close");

			send_headers_to(stdout, serv_hdr);
		}
		
		outlen = forward_content(serv_hdr, sockrfp, stdout);

		access_log (client_hdr, outlen, outlen, client_addr, giventime);
		return;
	}

	read_content(serv_hdr,sockrfp, &inbuf, &inlen);
	original_size = inlen;

	//if no data only forward header and exit
	if (inlen == 0) {
		logputs("Forwarding header only.");
		add_header(serv_hdr, "Connection: close");
		add_header(serv_hdr, "Proxy-Connection: close");

		send_headers_to(stdout, serv_hdr);
		fflush(stdout);

		access_log (client_hdr, 0, 0, client_addr, giventime);
		return;
	}
	
	//serv_hdr->chunklen isn't used anywhere else
	if(serv_hdr->chunklen > 0){
		if('1' == client_hdr->proto[7]){
			is_sending_data = 1;
			logputs("MaxSize reached - streaming original chunked data");
			add_header(serv_hdr, "Transfer-Encoding: chunked");
			add_header(serv_hdr, "Connection: close");
			add_header(serv_hdr, "Proxy-Connection: close");
			send_headers_to(stdout, serv_hdr);
			if(inlen > 0){
				printf("%X\r\n",inlen);//TODO verify format
				fwrite(inbuf,1,inlen,stdout);
				fputs("\r\n",stdout);
			}
			printf("%X\r\n",serv_hdr->chunklen);
			outlen = forward_content(serv_hdr, sockrfp, stdout);
		}else{
			//It is not worth to code proper unchunking
			//for this exceptional situation.
			send_error(500,"Internal error",(char*)0,
					"Too big file. Try using HTTP/1.1 client.");
		}

		access_log (client_hdr, outlen, outlen, client_addr, giventime);
		return;
	}
	
	if(inlen > MaxSize){
		is_sending_data = 1;
		logputs("MaxSize reached - streaming original data");
		add_header(serv_hdr, "Connection: close");
		add_header(serv_hdr, "Proxy-Connection: close");

		send_headers_to(stdout, serv_hdr);
		fwrite(inbuf,1,inlen,stdout);
		outlen = forward_content(serv_hdr, sockrfp, stdout);

		access_log (client_hdr, outlen, outlen, client_addr, giventime);
		return;
	}

	//in case something fails later and forgets to do this:
	outbuf = inbuf;
	outlen = inlen;

	if(inlen != serv_hdr->content_length) logprintf("In Content-Length: %d\n", inlen);

	logdifftime("Getting data");

	/* text/html optimizer */
	/* FIXME: inbuf must be at least (inlen + 1) chars big in order to hold added '\0' from htmlopt */
	if ((ProcessHTML) && (serv_hdr->type == TEXT_HTML) && (inlen > 0)) {
		HOPT_FLAGS hopt_flags = HOPT_NONE;

		if (ProcessHTML_CSS)
			hopt_flags |= HOPT_CSS;
		if (ProcessHTML_JS)
			hopt_flags |= HOPT_JAVASCRIPT;
		if (ProcessHTML_tags)
			hopt_flags |= HOPT_HTMLTAGS;
		if (ProcessHTML_text)
			hopt_flags |= HOPT_HTMLTEXT;
		if (ProcessHTML_PRE)
			hopt_flags |= HOPT_PRE;
		if (ProcessHTML_NoComments)
			hopt_flags |= HOPT_NOCOMMENTS;

		inlen = hopt_pack_html (inbuf, inlen, inbuf, hopt_flags);
		outlen = inlen;
	}

	/* text/css optimizer */
	/* FIXME: inbuf must be at least (inlen + 1) chars big in order to hold added '\0' from htmlopt */
	if ((ProcessCSS) && (serv_hdr->type == TEXT_CSS) && (inlen > 0)) {
		inlen = hopt_pack_css (inbuf, inlen, inbuf);
		outlen = inlen;
	}

	/* application/[x-]javascript optimizer */
	/* FIXME: inbuf must be at least (inlen + 1) chars big in order to hold added '\0' from htmlopt */
	if ((ProcessJS) && (serv_hdr->type == APPLICATION_JAVASCRIPT) && (inlen > 0)) {
		inlen = hopt_pack_javascript (inbuf, inlen, inbuf);
		outlen = inlen;
	}
	
	/* preemptive name resolution */
	if (PreemptNameRes){
		if ((serv_hdr->type == TEXT_HTML) && (inlen > 0))
			preempt_dns_from_html (inbuf, inlen);
	}

	if ((serv_hdr->type == IMG_PNG) || \
			(serv_hdr->type == IMG_GIF) || \
			(serv_hdr->type == IMG_JPEG) || \
			(serv_hdr->type == IMG_JP2K))
		status = compress_image(serv_hdr, client_hdr, inbuf, inlen, &outbuf, &outlen);

	if(serv_hdr->flags & DO_MODIFY){
		status = do_modify(serv_hdr, client_hdr, inbuf, inlen, &outbuf, &outlen,
				sockrfp, sockwfp);
		if(!status){
			inbuf = outbuf;
			inlen = outlen;
		} else {
			if (status == -1) {
				// TODO: unable to know the (possibly) compressed data size since it was already streamed
				access_log (client_hdr, original_size, -1, client_addr, giventime);
				exit (0);
			} else {
				exit (status);
			}
		}
	}

	if(serv_hdr->flags & DO_COMPRESS){
		status = do_compress(serv_hdr, inbuf, inlen, &outbuf, &outlen);
		if(!status){
			inbuf = outbuf;
			inlen = outlen;
		} else {
			if (status == -1) {
				// TODO: unable to know the compressed data size since it was already streamed
				access_log (client_hdr, original_size, -1, client_addr, giventime);
				exit (0);
			} else {
				exit (status);
			}
		}
	}
	
	logdifftime("Content modification/compression");
	logprintf("  and returned %d.\n", status);

	is_sending_data = 1;
	
	if (serv_hdr->flags & H_SIMPLE_RESPONSE ) {
		logputs("Sending HTTP/1.0 Simple Response.");
	}else{

		logputs("Forwarding header and modified content.");
		logputs("Out Headers:");

		snprintf(line, sizeof(line), "Content-Length: %d", outlen);
		if(serv_hdr->where_content_length > 0){
//				if(outlen > 0){
			serv_hdr->hdr[serv_hdr->where_content_length] = strdup(line);
//				}
		}else{
			add_header(serv_hdr, line);
		}
		add_header(serv_hdr, "Connection: close");
		add_header(serv_hdr, "Proxy-Connection: close");

		send_headers_to(stdout, serv_hdr);

	}

	//forward content
	tempp = outbuf;
	if (inlen != 0){
		int outcount = outlen;
		
		do{
			i = fwrite(tempp, 1, outcount, stdout);
			outcount -= i;
			tempp += i;
		}while((i > 0) && outcount);
		
		fflush(stdout);

		if(outcount == 0) {
			fsync(1);
			logdifftime("Forwarding");
		}
		else logprintf("Error - Last %d bytes of content could not be written\n",
			outcount);
	}
	access_log (client_hdr, original_size, outlen, client_addr, giventime);
}

void send_error( int status, char* title, char* extra_header, char* text )
    {
	    if(is_sending_data){
	    //if already sending data, sending error headers is pointless
		    status = -status;
	    }else{
		FILE *txtfile;
		char *txtfilename;
		char *txtfilebuf;
		int txtfilesize;
		int custom_error_sent = 0;
		
		switch (status){
		case 400:
			txtfilename = CustomError400;
			break;
		case 404:
			txtfilename = CustomError404;
			break;
		case 408:
			txtfilename = CustomError408;
			break;
		case 500:
			txtfilename = CustomError500;
			break;
		case 503:
			txtfilename = CustomError503;
			break;
		default:
			txtfilename = NULL;
		}

		/* displays custom error page, if specified and available */
		if (txtfilename != NULL){
			if (txtfile = fopen (txtfilename, "r")){
				fseek (txtfile, 0, SEEK_END);
				txtfilesize = ftell (txtfile);
				fseek (txtfile, 0, SEEK_SET);			
				if (txtfilebuf = malloc (txtfilesize + 1)){
					fread (txtfilebuf, 1, txtfilesize, txtfile);
					*(txtfilebuf+txtfilesize) = '\0';				
					send_headers( status, title, extra_header, "text/html", -1, -1 );
					printf ("%s", txtfilebuf); /* we don't want problems with '%' characters */
					fflush (stdout);
					custom_error_sent = 1;					
					free (txtfilebuf);
				}
				fclose (txtfile);
			}
		}
		
		if (custom_error_sent == 0){
			send_headers( status, title, extra_header, "text/html", -1, -1 );
			(void) printf( "<HTML><HEAD><TITLE>%d %s</TITLE></HEAD>\n<BODY BGCOLOR=\"#cc9999\"><H4>%d %s</H4>\n", status, title, status, title );
			(void) printf( "%s\n", text );
			(void) printf( "<HR>\n<ADDRESS><A HREF=\"%s\">%s</A></ADDRESS>\n</BODY></HTML>\n", ServUrl, SERVER_NAME );
			(void) fflush( stdout );
		}
	    }

    //log
    logprintf("ERROR - %d %s ( %s )\n", status, title, text);
    exit( 1 );
    }


void send_headers( int status, char* title, char* extra_header, char* mime_type, int length, time_t mod )
    {
    time_t now;
    char timebuf[100];

    (void) printf( "%s %d %s\r\n", PROTOCOL, status, title );
    (void) printf( "Server: %s\r\n", SERVER_NAME );
    now = time( (time_t*) 0 );
    (void) strftime( timebuf, sizeof(timebuf), RFC1123FMT, gmtime( &now ) );
    (void) printf( "Date: %s\r\n", timebuf );
    if ( extra_header != (char*) 0 )
	(void) printf( "%s\r\n", extra_header );
    if ( mime_type != (char*) 0 )
	(void) printf( "Content-Type: %s\r\n", mime_type );
    if ( length >= 0 )
	(void) printf( "Content-Length: %d\r\n", length );
    if ( mod != (time_t) -1 )
	{
	(void) strftime( timebuf, sizeof(timebuf), RFC1123FMT, gmtime( &mod ) );
	(void) printf( "Last-Modified: %s\r\n", timebuf );
	}
    (void) printf( "Connection: close\r\nProxy-Connection: close\r\n\r\n" );
    }


http_headers *new_headers(void){
	http_headers *h = malloc(sizeof(http_headers));

	/* FIXME: This is a temporary fix,
	 * it prevents eventual crashing */
	memset (h, 0, sizeof(http_headers));
	
	h->lines = h->flags = 0;
	h->where_content_type = h->where_content_length = 
		h->where_chunked = h->where_content_encoding = 
		h->where_etag = h->port = h->status = 
		h->content_length = h->chunklen -1;
	
	h->method = h->url = h->path = h->host = h->proto = (char*)0;
	
	return h;
}

char *find_header(char* key, http_headers *hdr){
	int i;
	for(i=0; i< hdr->lines; i++)
		if(!strncasecmp(hdr->hdr[i], key, strlen(key)))
			return hdr->hdr[i] + strlen(key);
	return (char*)0;
}

/*Returns index in array of headers*/
int add_header(http_headers *hdr, char *newhdr){
	if(hdr->lines + 1 <MAX_HEADERS) {
		hdr->hdr[hdr->lines] = strdup(newhdr);
		hdr->lines ++;
	}
	else {
		char *keyword;
		
		if(hdr->status == -1)
			keyword = "Bad request";
		else 
			keyword = "Bad response";
		
		send_error( 400, keyword , (char*) 0, "Too many headers.");
	}
	return hdr->lines - 1;
}


int remove_header(http_headers *hdr, int n){
	int i;
	if(n >= hdr->lines || n < 0) return -1;

	for(i=n; i<hdr->lines - 1;i++){
		hdr->hdr[i] = hdr->hdr[i+1];
	}
	hdr->hdr[i] = (char*)0;
	hdr->lines--;
	return 0;
}
	
http_headers * parse_initial_request(void){
	size_t linelen;
	int was_auth=0,i=0;

	http_headers *hdr = new_headers();
	/* Read the first line of the request. */
	if (fgets(line, sizeof(line), stdin ) ==  NULL)
	{
		send_error( 400, "Bad Request", (char*) 0, 
					"No request found or request too long." );
	}
	linelen = check_trim(line);
	if (linelen == 0)
	{
                send_error( 400, "Bad Request", (char*) 0,
					"No request found or request too long." );
	}
	add_header(hdr, line);

	hdr->method=(char*)malloc(5*linelen);
	hdr->url = hdr->method + linelen;
	hdr->proto = hdr->url + linelen;
	hdr->host = hdr->proto + linelen;
	hdr->path = hdr->host + linelen;

	/* Parse it. */
	if ( sscanf( line, "%[^ ] %[^ ] %[^ ]", hdr->method, hdr->url, hdr->proto ) != 3 )
		send_error( 400, "Bad Request", (char*) 0, "Can't parse request." );

	if ((char*) 0 == hdr->url)
		send_error( 400, "Bad Request", (char*) 0, "Null URL." );

	if ( strncasecmp( hdr->url, "http://", 7 ) == 0 )
	{

		strncpy( hdr->url, "http", 4 );	/* make sure it's lower case */
		if ( sscanf( hdr->url, "http://%[^:/]:%hd%s", hdr->host, &hdr->port, hdr->path ) == 3 );
		else if ( sscanf( hdr->url, "http://%[^/]%s", hdr->host, hdr->path ) == 2 )
		{
		   	hdr->port = 80;
		}
		else if ( sscanf( hdr->url, "http://%[^:/]:%hd", hdr->host, &hdr->port ) == 2 )
		{
			hdr->path=hdr->url;
		}
		else if ( sscanf( hdr->url, "http://%[^/]", hdr->host ) == 1 )
		{
		hdr->port = 80;
			hdr->path=hdr->url;
		}
		else
			send_error( 400, "Bad Request", (char*) 0, "Can't parse URL." );

		if(NextProxy)	hdr->path = hdr->url;
		//log URL
		logprintf("URL - %s\n", hdr->url);

		if(ModifySuffixes && undo_name(hdr->path)) hdr->flags |= H_SUFF_MODIFIED;
	}
	else if ( strcmp( hdr->method, "CONNECT" ) == 0 )
	{
		if ( sscanf( hdr->url, "%[^:]:%hd", hdr->host, &hdr->port ) == 2 ) ;
		else if ( sscanf( hdr->url, "%s", hdr->host ) == 1 )//equivalent to strcpy until first whitespace?
		    hdr->port = 443;
		else
		    send_error( 400, "Bad Request", (char*) 0, "Can't parse URL." );

		if (PasswdFile) {
		    // Look for auth header
		    while (fgets(line, sizeof(line), stdin) != NULL)
		    {
			if (strcmp(line, "\n") == 0 || strcmp(line, "\r\n") == 0)
			    break;
			    
			if (strncasecmp(line,"Proxy-Authorization: Basic ",27) == 0 && strlen(line) > 30) {
			    for (i=0; i<PasswdNumEntries; i++ ) {
				if (strlen(PasswdEntries[i]) == strlen (base64_decode(&(line[27]))) && strncmp( base64_decode(&(line[27])), PasswdEntries[i], strlen(PasswdEntries[i]) ) == 0 ) {
				    was_auth=1;
				}
			    }
			}
		    }

		    if(!was_auth) {
			logputs("Requesting HTTP auth from client for CONNECT method");
			send_error( 407, "Proxy Authentication Required", "Proxy-Authenticate: Basic realm=", "You have to be registered before using this proxy.");
		    }

		}

		hdr->flags |= H_USE_SSL;
		logprintf("URL - %s\n", hdr->url);
	}
	else if ((*(hdr->url) == '/') && TransparentProxy)
	{
		hdr->flags |= H_TRANSP_PROXY_REQUEST;

		hdr->port = 80;
		strcpy (hdr->path, hdr->url);
		*(hdr->host) = '\0';	// hdr->host is left undefined for now, we need Host: data at a later stage
		/* hdr->url will need to be fixed at a later stage (at this point it only has the file path) */

		if(NextProxy)	hdr->path = hdr->url;

		//log URL
		// FIXME: this will not log the URL, but only the path instead (no hostname)
		logprintf("URL - %s\n", hdr->url);

		if(ModifySuffixes && undo_name(hdr->path)) hdr->flags |= H_SUFF_MODIFIED;
	}
	else
		send_error( 400, "Bad Request", (char*) 0, "Unknown URL type." );
		
	return hdr;
}

//If it returns line length > 0, line was correctly ended with newline - use to 
//check output
//from fgets(). 
static int check_trim( char* line ) {
    int k,l;

    k = l = strlen( line );
	
    while (l > 0 && (line[l-1] == '\n' || line[l-1] == '\r'))
		line[--l] = '\0';

	return (k != l)? l : 0;
}


void get_client_headers(http_headers * hdr){
	int was_via=0, was_keepalive=0;
	int linelen;
	int was_auth=0;
	int unsigned long i;
	
	logputs("Headers from client:");
	
		
	while (fgets(line, sizeof(line), stdin) != NULL)
	{
		if (strcmp(line, "\n") == 0 || strcmp(line, "\r\n") == 0)
			break;
		
		if((linelen=check_trim(line))==0)
			send_error( 400, "Bad request", (char*) 0, "Line too long.");

//		if (strncasecmp(line,"User-Agent:", 11) == 0)
		logputs_hdr(line);
		clean_hdr(line);
		if (strncasecmp(line,"Connection:",11) == 0)
				continue;

		if (strncasecmp(line,"Proxy-Connection:",17) == 0)
				continue;

		if (strncasecmp(line,"Proxy-Authorization:",20) == 0 && strlen(line) > 30) {
		    for (i=0; i<PasswdNumEntries; i++ ) {
			if (strlen(PasswdEntries[i]) == strlen (base64_decode(&(line[27]))) && strncmp( base64_decode(&(line[27])), PasswdEntries[i], strlen(PasswdEntries[i]) ) == 0 ) {
				was_auth=1;
				continue;
			}
		    }
		}

		if (strncasecmp(line,"Keep-Alive:",11) == 0)
				continue;
	
		//hmmm, this should not happen. But we should forward this unchanged
		//instead!
		if (strncasecmp(line, "Range:", 6) == 0) continue; 

		// define host data if empty
		if (strncasecmp(line,"Host:", 5) == 0) {
			if (*(hdr->host) == '\0') {
				char *new_host, *provided_host_str;
				
				provided_host_str = line + 5;
				if (*provided_host_str == ' ')
					provided_host_str++;
				
				new_host = malloc (strlen (provided_host_str) + 1);
				strcpy (new_host, provided_host_str);
				hdr->host = new_host;
			}
		}
		
		//get content length
		if (strncasecmp(line, "Content-Length:", 15 ) == 0)
			hdr->content_length = atol(&(line[15]));

			//can accept gzip?
		else if (strncasecmp(line, "Accept-Encoding:", 16) == 0)
			//we don't know how to uncompress HTML 
			//to modify it -> don't pass this to server
		{
			if(DoGzip) hdr->flags |= H_WILLGZIP;
			continue; 
		}

		if(strncasecmp(line,"Via:",4) == 0){
			if (ServHost) {
				if(strstr(line,ServHost) && strstr(line, SERVER_NAME)) 
						/*prevent infinite recursion*/
					send_error( 503, "Service Unavailable", (char*) 0,
							"Connection refused (based on Via header)." );
				
				if(linelen + strlen(ServHost) + 
					sizeof(SERVER_NAME) + 15 >= MAX_LINELEN) continue;
				
				snprintf(line + linelen , sizeof(line) - linelen, 
						", 1.1 %s (%s)",ServHost ,SERVER_NAME);
				was_via = 1;
				
			}else continue; //forget Via header
		}
		add_header(hdr, line);
		
	}

	if(ServHost && !was_via)
	{
		snprintf(line,sizeof(line), "Via: 1.1 %s (%s)",ServHost , SERVER_NAME);
		add_header(hdr, line);
	}

	if(PasswdFile && !was_auth) {
		logputs("Requesting HTTP auth from client");
		send_error( 407, "Proxy Authentication Required", "Proxy-Authenticate: Basic realm=", "You have to be registered before using this proxy.");
	}
}//get_client_headers

/*Does not release the headers ! Rather use special function?*/
void send_headers_to(FILE * sockfp, http_headers *hdr){
	int i, len;
	
	if (0 == hdr->lines) return; //useful for simple response
	
	if (hdr->status == -1){//treat first line specially
		i = 1;
		len = snprintf(line,sizeof(line),"%s %s %s\r\n", hdr->method, hdr->path, hdr->proto);
		fputs(line, sockfp);
		
		line[len-2] = '\0';
		logputs_hdr(line);
	}else i = 0;
		
	// Forward the remainder of the request from the client
	for(;i < hdr->lines && i<MAX_HEADERS;i++)
	{
		fputs(hdr->hdr[i],sockfp);		
		logputs_hdr(hdr->hdr[i]);
		fputs("\r\n",sockfp);
	}
	fputs("\r\n",sockfp);
	fflush(sockfp);
}

/* Use faster (thread-unsafe) variants of getc() and putc(), if possible.
 */
#ifdef getc_unlocked
#define fastgetc(x) getc_unlocked(x)
#else
#define fastgetc(x) getc(x)
#endif
		
#ifdef putc_unlocked
#define fastputc(x,y) putc_unlocked(x,y)
#else
#define fastputc(x,y) putc(x,y)
#endif

// returns forwarded content size
int forward_content(http_headers *hdr, FILE *from, FILE *to){
	int i, ch;
	for (i = 0; (hdr->content_length == -1 || i < hdr->content_length) && 
		(ch = fastgetc(from)) != EOF; ++i) {
	
		fastputc(ch, to);
		// If we are sending a big file down a slow line, we
		// need to reset the alarm once a while.
		if (ZTimeout && (i%10000 == 0))
		  alarm(ZTimeout);
	}
	fflush(to);

	return (i);
}

	//get data and store for compression
static void read_content(http_headers *hdr, FILE *sockrfp, char ** inbuf, int *inlen){
	int i, inbuf_start = 0;
	unsigned int chunklen;
	unsigned int chunkpos;
	int ich = EOF;

	if (hdr->content_length == -1) *inlen = 1024;
	else *inlen = hdr->content_length;

	*inbuf=malloc(*inlen + 1);

	if (hdr->flags & H_SIMPLE_RESPONSE) {
	  strncpy((*inbuf), line, 4);
	  inbuf_start = 4;
	}

	// unchunk if needed
	if (hdr->where_chunked > 0)
	{
		remove_header(hdr, hdr->where_chunked);
		hdr->where_chunked = -1;
		//chunked
		chunkpos = 0;
		fscanf(sockrfp, "%x", &chunklen);
		//Eat any chunk-extension(RFC2616) up to CRLF.
		while(!((fastgetc(sockrfp) == '\r') && (fastgetc(sockrfp) == '\n')));
		while (chunklen > 0)
		{
			if (MaxSize && (chunklen + chunkpos > MaxSize)){
				*inlen = chunkpos;
				hdr->chunklen = chunklen;
				return;
			}
			for (i = 0; i < chunklen; i++){
				
				if (chunkpos + i >= *inlen)
				{
					(*inbuf)=realloc((*inbuf),2* (*inlen) + 1);
					(*inlen) *=2;
				}
				ich = fastgetc(sockrfp);
				if (ich == EOF) break;
				
				(*inbuf)[chunkpos + i] = ich;
			}
			chunkpos += i;

			if (ich == EOF) break;

			chunklen = 0;
			if((fastgetc(sockrfp)!= '\r') || 
					(fastgetc(sockrfp) != '\n')){
				//TODO Here we probably should forward 
				//everything to client to let him deal with the 
				//mess. This requires backup buffer where 
				//original response would be stored.
				send_error(500, "Internal Error",(char*)0,
						"Reading of chunked encoding failed. " 
						"Sorry, either the server or ziproxy is buggy.");
			}
			fscanf(sockrfp, "%x", &chunklen);
			//Eat any chunk-extension(RFC2616) up to CRLF.
			while(!((fastgetc(sockrfp) == '\r') && (fastgetc(sockrfp) == '\n')));
		}

		if(0==chunklen){//assume empty trailer
			fgetc(sockrfp); fgetc(sockrfp);
			
		}
		*inlen = chunkpos;
	} else { //not chunked
		if (hdr->content_length == -1)
		{
			i = inbuf_start;
			while ((ich = fastgetc(sockrfp)) != EOF)
			{
				if (i >= *inlen)
				{
					(*inbuf)=realloc((*inbuf),2*(*inlen) + 1);
					(*inlen) *= 2;
				}
				
				(*inbuf)[i] = ich;
				i++;

				if (i > MaxSize)
					break;

			}
			*inlen = i;
		}
		else 
		{
			for (i = 0; i < hdr->content_length; i++)
			{
				ich = fastgetc(sockrfp);
				
				if(ich == EOF) break;
				
				(*inbuf)[i] = ich;
			} 
			*inlen = i;
		}
	}
}
		
#undef fastgetc
#undef fastputc

http_headers * get_response_headers(FILE *sockrfp){
	http_headers *hdr = new_headers();
	int n, linelen, *savepos;
	char *tempp;
	
	// Process the first few characters to see if it's an HTTP/1.0
	// simple response i.e. a response without any header
	// lines.

	// read the first 4 characters
	fgets(line, 5, sockrfp);
	n = strlen(line);
	if (0 == n){
		send_error(500, "Server error", NULL, "Empty response from server");
	}
	
	if (n < 4 || strncasecmp(line, "HTTP", 4)) {
		hdr->flags |= H_SIMPLE_RESPONSE; // It's a simple response.
		// Check for image signatures
		hdr->type = detect_type(line, 4);
		hdr->lines = 0;
		//cache the line here
		hdr->hdr[0] = strdup(line);

		logputs("Received HTTP/1.0 Simple Response.");
		// We're passing along all other types of data.
		// TODO: Handle other compressable types.
		return hdr;
	}
	// It's a full response

	// read the rest of the first line
	fgets(line+n, sizeof(line), sockrfp);

	//get header
	do{
		if (strcmp(line, "\n") == 0 || strcmp(line, "\r\n") == 0)
			break;

		if ((linelen = check_trim(line)) == 0)
				send_error(500,"Internal Error",NULL,
				"Too long line in response from server");
		
		logputs_hdr(line);
		clean_hdr(line);
		savepos = (int*)0;
		
		//get content length
		if (strncasecmp(line, "Content-Length:", 15) == 0)
		{
			hdr->content_length = atol(&(line[15]));
			savepos = &hdr->where_content_length;
		}
		else if (strncasecmp(line, "Content-Encoding:", 17) == 0)
			savepos = &hdr->where_content_encoding;

		else if (strncasecmp(line, "Transfer-Encoding:", 18) == 0)
		{//check for chunking
			tempp = line + 19;
			if (strncasecmp(tempp, "chunked", 7) == 0)
			savepos = &hdr->where_chunked;
		}

		else if (strncasecmp(line, "Content-Type:", 13) == 0)
			savepos = &hdr->where_content_type;

		else if (ModifySuffixes && (strncasecmp(line, "Location:", 9) == 0))
		{//change suffix in Location: header
			tempp = strrchr(line, '.');
			if(tempp && (tempp + 12 < line + MAX_LINELEN) &&  
				(!strcasecmp(tempp,".gif") || !strcasecmp(tempp,".png")))
				
				strcpy(tempp + 4, "_or_jpg");
		}
		else if (strncasecmp(line, "Accept-Ranges:", 14) == 0)
				strcpy(&line[14], " none");
				
		else if (strncasecmp(line, "ETag:", 5) == 0)
			savepos = &hdr->where_etag;

		else if ((strncasecmp(line, "HTTP/1.1", 8) == 0) ||
			(strncasecmp(line, "HTTP/1.0", 8) == 0))
			hdr->status = atoi(&line[8]);
		else{
			tempp = (char*)0;

			if(!strncasecmp(line, "Connection:", 11))
					tempp = line + 11;
			else if(!strncasecmp(line, "Proxy-Connection:", 17))
					tempp = line + 17;
			else if(!strncasecmp(line,"Keep-Alive:",11))
					tempp = line;
			
			if(tempp && strstr(tempp, "Keep-Alive")) 
				hdr->flags |= H_KEEPALIVE;
		}
		//store header info
		if(strncasecmp(line, "Connection:", 11) && 
		 strncasecmp(line, "Proxy-Connection:", 17) &&
		 strncasecmp(line,"Keep-Alive:",11)){
			n = add_header(hdr,line);
			if(savepos) *savepos = n;
			if(0 == n) hdr->proto = hdr->hdr[0]; 
		}

	} while (fgets(line, sizeof(line), sockrfp) != 0);
	return hdr;
}
/*
 * chdr - headers of client request
 * shdr - headers of server response
 */

void decide_what_to_do(http_headers *chdr, http_headers *shdr){
	char *tempp;
	int i;

	shdr->type = OTHER_CONTENT;
	shdr->flags &= ~DO_COMPRESS;
	shdr->flags &= ~DO_MODIFY;
	
	if(-1 == shdr->where_content_type) return; 
	
	tempp = shdr->hdr[shdr->where_content_type] + 14;
	if (!strncasecmp(tempp,"text/html", 9)) {
		shdr->type = TEXT_HTML;
		if(DoGzip) shdr->flags |= DO_COMPRESS;
		if(ModifySuffixes) shdr->flags |= DO_MODIFY;
	} else if (!strncasecmp(tempp,"text/css", 8)) {
		shdr->type = TEXT_CSS;
		if(DoGzip) shdr->flags |= DO_COMPRESS;
	} else if (!strncasecmp(tempp,"text", 4)) {
		if(DoGzip) shdr->flags |= DO_COMPRESS;
	} else if (ProcessJPG &&
		(!strncasecmp(tempp, "image/jpeg", 10) ||
		!strncasecmp(tempp, "image/pjpeg", 11) ||
		!strncasecmp(tempp, "image/jpg", 9) ||
		!strncasecmp(tempp, "image/pjpg", 10) 
		))
		 shdr->type = IMG_JPEG;
	else if (ProcessGIF && !strncasecmp(tempp, "image/gif", 9))
		shdr->type = IMG_GIF;
	else if (ProcessPNG && !strncasecmp(tempp, "image/png", 9))
		shdr->type = IMG_PNG;
	else if ((!strncasecmp(tempp,"application/x-javascript", 24)) || (!strncasecmp(tempp,"application/javascript", 22))) {
		shdr->type = APPLICATION_JAVASCRIPT;
		if(DoGzip) shdr->flags |= DO_COMPRESS;
	//Check compressible types under "application/"
	} else if(DoGzip && Compressible &&
			(strncasecmp(tempp, "application/",12)==0) && 
			(strlen(tempp) > 12))
	{
	tempp += 12;
		if(strncasecmp(tempp,"x-",2) == 0) tempp+=2;

		for (i=0; Compressible[i]; i++)
			if (strncasecmp(tempp,Compressible[i],sizeof(Compressible[i]) - 1)==0)
			{
				shdr->flags |= DO_COMPRESS;
				break;
			}
	}

#ifndef JP2K
	if(ModifySuffixes && (shdr->type != OTHER_CONTENT) && !(chdr->flags & H_SUFF_MODIFIED))
	{
			tempp = strrchr(chdr->path,'.');
			if(tempp && (!strcasecmp(tempp,".gif") || !strcasecmp(tempp,".png")))
					shdr->type = OTHER_CONTENT;
			//we cease to do anything with image. It may confuse 
			//browser if content type is other than file suffix.
	}
#endif
	
	//Dont gzip, if either:
	//  content is encoded (e.g. already compressed)
	//  browser doesn't accept gzip
	if((shdr->where_content_encoding > 0) || !(chdr->flags & H_WILLGZIP))
		shdr->flags &= ~DO_COMPRESS;
	
	//Dont modify HTMLs if content is encoded
	if(shdr->where_content_encoding > 0)
		shdr->flags &= ~DO_MODIFY;
}

//Remove extra whitespace that may prevent correct parsing.
void clean_hdr(char* ln){
	int spaces = 0;
	char * colon, *i;
	
	colon = strchr(ln,':');
	if(! colon) return;
	for(i = colon + 1; *i != '\0'; i++){
		if(' ' == *i )
			spaces++;
		else break;//*i is now first char after space(s)
	}
	if (spaces > 1) memmove(colon+2,i,strlen(ln)-(colon-ln) - 2);
}

// Decode login:pass pair for HTTP authorization
char *base64_decode(char *input) {
    unsigned char in[4], out[3], v;
    int r, i, len, str_len;
    char *outstr;
    int outpos=0;

    outstr=(char *)malloc(512);
    str_len = strlen(input);

    r = 0;
    while (input[r] && outpos < 508 ) {
        for (len = 0, i=0; i<4 && input[r]; i++) {
            v = 0;
            while(input[r] && v == 0 ) {
                v = (unsigned char) input[r++];
                v = (unsigned char) ((v < 43 || v > 122) ? 0 : cd64[ v - 43 ]);
                if(v)
		  v = (unsigned char) ((v == '$') ? 0 : v - 61);
            }
            if (input[r] || ( (str_len % 4) == 0 && input[str_len-1] != '=' ) ) {
	      len++;
	      if (v)
		in[i] = (unsigned char) (v - 1);
	    } else {
	      in[i] = 0;
	    }
        }

        if(len) {
	  out[0] = (unsigned char) (in[0] << 2 | in[1] >> 4);
	  out[1] = (unsigned char) (in[1] << 4 | in[2] >> 2);
	  out[2] = (unsigned char) (((in[2] << 6) & 0xc0) | in[3]);
	  for (i=0; i<len-1; i++)
	    outstr[outpos++] = out[i];

        }
    }
    outstr[outpos]=0;
    return(outstr);
}

/*
 * transform the request location from HTTP-host (path) to proxy format (full URL), for internal Ziproxy's usage
 * it's used when processing requests as transparent proxy
 * -- requires get_response_headers() being called previously
 */
void fix_request_url (http_headers *hdr) {
	char *new_url;
	
	if (*(hdr->host) == '\0')
		send_error (400, "Bad Request", (char*) 0, "Malformed request or non-HTTP/1.1 compliant.");

	new_url = malloc (strlen (hdr->url) + strlen (hdr->host) + 7 + 1);
	sprintf (new_url, "http://%s%s", hdr->host, hdr->url);
	hdr->url = new_url;
}

