/* http.h 
 * Part of ziproxy package.
 * 
 * Copyright (c)2003-2004 Juraj Variny<variny@naex.sk>
 * Copyright (c)2005-2007 Daniel Mealha Cabrita
 *
 * Released subject to GNU General Public License.
 * 
 * HTTP communication functions.
 */

#include <stdio.h>
#include <time.h>
#include <sys/time.h>
//If this is included in http.c, make declarations local.
#ifdef SRC_HTTP_C
#define EXTERN
#else
#define EXTERN extern
#endif 

//To stop multiple inclusions.
#ifndef SRC_HTTP_H
#define SRC_HTTP_H

#define MAX_HEADERS 200


typedef struct {
	char * hdr[MAX_HEADERS];
	int lines;
	
	int where_content_type, where_content_length, where_chunked;
	int where_content_encoding, where_etag;

	//If headers contain request from client, status is always -1.
	int status;

	char *method, *url, *proto, *host, *path;
	int port, content_length, flags, type;

	int chunklen;	
} http_headers;

#define H_WILLGZIP (1<<1)
#define H_SUFF_MODIFIED (1<<2)
#define H_USE_SSL (1<<3)
#define H_KEEPALIVE (1<<4)
#define H_SIMPLE_RESPONSE (1<<5)
#define H_TRANSP_PROXY_REQUEST (1<<6)

#define DO_COMPRESS (1<<10)
#define DO_MODIFY (1<<11)


EXTERN int is_sending_data;

EXTERN http_headers * parse_initial_request(void);
EXTERN void proxy_http (http_headers *client_hdr, FILE* sockrfp, FILE* sockwfp, const char *client_addr, struct timeval *giventime);

EXTERN void send_error( int status, char* title, char* extra_header, char* text );
EXTERN void send_headers( int status, char* title, char* extra_header, char* mime_type, int length, time_t mod );

EXTERN http_headers *new_headers(void);
EXTERN char *find_header(char* key, http_headers *hdr);
EXTERN int add_header(http_headers *hdr, char *newhdr);
EXTERN int remove_header(http_headers *hdr, int n);
EXTERN http_headers * get_response_headers(FILE *sockrfp);
EXTERN void send_headers_to(FILE * sockfp, http_headers *hdr);
EXTERN void decide_what_to_do(http_headers *chdr, http_headers *shdr);

EXTERN void fix_request_url (http_headers *hdr);
EXTERN void get_client_headers(http_headers * hdr);

#define SERVER_NAME "ziproxy"
#define PROTOCOL "HTTP/1.1"
#define RFC1123FMT "%a, %d %b %Y %H:%M:%S GMT"
#define MAX_LINELEN 16384

#endif //SRC_HTTP_H

#undef EXTERN

