/* text.c - part of ziproxy package
 *
 * Copyright (c)2003-2004 Juraj Variny<variny@naex.sk>
 * Copyright (c)2005-2006 Daniel Mealha Cabrita
 *
 * Released subject to GNU General Public License
 *
 * HTML modification, text compression fuctions
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h> //for off_t
#include <zlib.h>
#include <assert.h>

#include "http.h"
#include "cfgfile.h"
#include "htmlmodify.h"
#include "image.h"

#define CHUNKSIZE 4050

static int gzip(char* inbuf, int inlen, char** outbuf, int *outlen, int out_type);

#ifdef JP2K
//JP2 version will further communicate with server.
static http_headers *shdr;
static http_headers *chdr;
static FILE *to_sock;
static FILE *from_sock;

#endif

enum {ONormal,OChunked, OStream, OGzipStream};

/* returns:
 *   0 ok,
 *  -1 ok, data streamed already
 *  >0 ERROR
 */
int do_modify(http_headers *serv_hdr, http_headers *client_hdr, char *inbuf, int insize, 
		char ** outbuf, int *outsize, FILE *sockrfp, FILE *sockwfp){
	 int status;

#ifdef JP2K
	 to_sock = sockwfp;
	 from_sock = sockrfp;
	 shdr = serv_hdr;
	 chdr = client_hdr;
#endif

	inbuf[insize]='\0';
	if((insize > MinTextStream) && (('1' == client_hdr->proto[7]) || !UseContentLength))
	{
		int out_type;
		is_sending_data = 1;
		if(('1' == client_hdr->proto[7]) && !(serv_hdr->flags & DO_COMPRESS)){ //use chunked
		//some browsers doesn't like chunked with HTTP/1.0
			if('0' == serv_hdr->hdr[0][7]) serv_hdr->hdr[0][7]='1';
			out_type = OChunked;		
		}else if(serv_hdr->flags & DO_COMPRESS)
			out_type = OGzipStream;
			
		else out_type = OStream;
				
		if(out_type == OChunked) add_header(serv_hdr, "Transfer-Encoding: chunked");
		if(out_type == OGzipStream) add_header(serv_hdr, "Content-Encoding: gzip");

		add_header(serv_hdr, "Connection: close");
		add_header(serv_hdr, "Proxy-Connection: close");

		logputs("While modification, streaming "
				"content. Out Headers:");
		send_headers_to(stdout, serv_hdr);

		status = htmlmodify(inbuf,outbuf, out_type);
		
		logdifftime("Modification+streaming");
		if (!status)
			return (-1);
		else
			return (status);
	}else{
		status = htmlmodify(inbuf,outbuf, ONormal);
		*outsize = strlen(*outbuf);
		return (status);
	}

}

//TODO correct return value, print status into logs
/* returns:
 *   0 ok,
 *  -1 ok, data streamed already
 *  >0 ERROR
 */
int do_compress(http_headers *hdr, char *inbuf, int insize, 
		char ** outbuf, int *outsize){
	int status;
	
	is_sending_data = 1;
	if(UseContentLength || (insize < MinTextStream)) {
		status = gzip (inbuf, insize, outbuf, outsize, ONormal);

		if (!status)
			add_header(hdr, "Content-Encoding: gzip");
		return (status);
	}

	add_header(hdr, "Content-Encoding: gzip");
	add_header(hdr, "Connection: close");
	add_header(hdr, "Proxy-Connection: close");
	
	logputs("While compression, streaming "
			"content. Out Headers:");
	remove_header(hdr, hdr->where_content_length);
        hdr->where_content_length=-1;
	send_headers_to(stdout, hdr);

	status = gzip(inbuf, insize, outbuf, outsize, OGzipStream);

	logdifftime("Compression+streaming");
	if (!status)
		return (-1);
	else
		return (status);
}


int undo_name (char *path)
{
	char * tempp;

	tempp = strrchr(path,'.');
#ifdef JP2K
	if(tempp && (!strcasecmp(tempp,suff)))
	{
		*tempp=0;
		return 1;
	}

#else
	if(tempp && (!strcasecmp(tempp,".gif_or_jpg") ||
		!strcasecmp(tempp,".png_or_jpg")))
	{
		tempp[4]=0;
		return 1;
	}
#endif		
	return 0;
}



static int gzip(char* inbuf, int inlen, char** outbuf, int *outlen, int out_type)
{
		int filedes = -1;
		char filenam[] = "/tmp/ziproxy_XXXXXX";
		gzFile gzfile = 0;
		
		if(out_type == ONormal){
			if((filedes = mkstemp(filenam)) < 0) return 10;
			unlink(filenam);
			if(!(gzfile = gzdopen(dup(filedes), "wb9"))) return 20;
			}
		else if(out_type == OGzipStream){
			if(!(gzfile = gzdopen(dup(1), "wb9"))) return 20;
			}
		else assert("Invalid out_type parameter to gzip()!\n");
			
		if(!(gzwrite(gzfile, inbuf, inlen)))
		{
				gzclose(gzfile);
				return 30;
		}
		
		if(gzclose(gzfile) != Z_OK) return 40;

		free(inbuf);
		
		if(out_type == ONormal){
			*outlen=lseek(filedes,0,SEEK_END);
			*outbuf = malloc(*outlen);
		
			lseek(filedes,0,SEEK_SET);

			//read gz file
			read(filedes,*outbuf, *outlen);
		
			close(filedes);
		}
		return 0;
}

#ifdef JP2K

//IMG tags locations already processed.
typedef struct struct_imglist{
	char *location;
	int converted;
	struct struct_imglist *next;
}imglist;

static int find_img(imglist *first,char *loc){
	imglist *nextl = first;
	while(nextl){
		if(!strcmp(nextl->location, loc))
			return nextl->converted;
		nextl = nextl->next;
	}
	
	return -1;
}

static void add_location(imglist **first, char *loc, int converted){
	imglist *newl = (imglist*)malloc(sizeof(imglist));
	newl->location = loc;
	newl->converted = converted;
	newl->next = *first;
	*first = newl;
}

char * IMGtoOBJ(char *src, char *original, int attrs, char ** img_attr, char ** img_val){
	char *ret;
	char *w = 0, *h = 0, *alt = 0, *tmp;
	char line[MAX_LINELEN];
	int i;
	long width, height;
	float rate;  
	static imglist *first =(imglist *)0;
	http_headers *request, *response = (http_headers*)0;

	for(i=0;i<attrs;i++){//find useful attributes
		if(!strcasecmp(img_attr[i], "width"))
			w = img_val[i];
		else if(!strcasecmp(img_attr[i], "height"))
			h = img_val[i];
		else if(!strcasecmp(img_attr[i], "alt"))	
			alt = img_val[i];
	}

	if(!w || !h)//width & height MUST be known
		return original;
	i = find_img(first, src);
	
	if(0 == i) return original;
	if(-1 == i){

		width = strtol(w,&tmp,10);
		if(tmp == w || width < 0) return original;
		if('%'== *tmp) width *= 8; //assume 800x600 screen

		height = strtol(h,&tmp,10);
		if(tmp == h || height < 0) return original;
		if('%'== *tmp) height *= 6; //assume 800x600 screen
		
		//Ask server for image size. If image is on other host (and we aren't 
		//using proxy), refrain.
		if((shdr->flags & H_KEEPALIVE) && (NextProxy || strncasecmp(src, "http://", 7))){
			request = new_headers();
			snprintf(line, sizeof(line), "HEAD %s HTTP/1.1", src);
			request->hdr[0] = strdup(line);
			request->method = "HEAD";
			request->path = src;
			request->proto = PROTOCOL;
			
			snprintf(line, sizeof(line), "Host: %s", chdr->host);
			request->hdr[1] = strdup(line);
			request->hdr[2] = "Connection: Keep-Alive";
			request->hdr[3] = "Keep-Alive: 300";
			request->lines = 4;
			logprintf("Requesting size of '%s':",src);
			send_headers_to(to_sock, request);
			logputs ("Response:");
			response = get_response_headers(from_sock);
			if(response->status != 200) response->content_length = -1;
		}

		getRateQuality(width, height, &i, &rate);
		if(rate < 0.0001){
			add_location(&first, src, 0);
			return original;
		}
		//check if it is image type we can't handle	
		if(response->where_content_type > 0){
			request->flags |= H_SUFF_MODIFIED;
			decide_what_to_do(request, response);
			if((response->type != IMG_PNG) &&
				(response->type != IMG_GIF) &&
				(response->type != IMG_JPEG) &&
				(response->type != IMG_JP2K)) {
				add_location(&first, src, 0);
				return original;
			}
		}
	
		if(response && response->content_length > 0){
			//we have all information we want
			
			i = response->content_length * 0.85 + 30;
			
			if(rate*width*height*3 > i){
				add_location(&first, src, 0);
				return original;
			}else add_location(&first, src, 1);
		}
	}

	if(alt)
		snprintf(line, sizeof(line), "<object TYPE=\"image/jp2\" WIDTH=\"%s\" HEIGHT=\"%s\" "\
		"STANDBY=\"%s\"><param name=\"SRC\" value=\"%s%s\"> %s </object>",
		w, h, alt, src, suff, original);
	else
		snprintf(line, sizeof(line), "<object TYPE=\"image/jp2\" WIDTH=\"%s\" HEIGHT=\"%s\" >"\
		"<param name=\"SRC\" value=\"%s%s\"> %s </object>", w, h, src, suff, 
		original);

	ret = strdup(line);
	
	
/*	asprintf(&ret, "<embed type='image/jp2' src = \"%s%s\" %s ></embed>", 
src, suff, wh );
*/
	return ret;
}
#endif
